#
# This awk program converts Adobe font metric (afm) files to TeX
# "pl" (font property list) files.  It is intended to be used in
# conjunction with pltotf to create a .tfm file for one of the
# internal fonts in a PostScript printer.
#
#  -- Copyright (c) 1987 by Van Jacobson, Lawrence Berkeley Lab.
#
# $Header: af2pl.awk,v 1.5 87/12/04 15:22:01 van Locked $

BEGIN			{
			# constant to convert AFM units to TeX
			# tfm units.
			afmscale = 72.27 / 72;
			printf "(DESIGNSIZE R %g)\n", 10*afmscale
			print "(COMMENT DesignSize is in points)"
			}
/^Comment /		{
			printf "(COMMENT"
			for (i = 1; ++i <= NF; )
				printf " %s",$i
			printf ")\n"
			}
/^Notice /		{
			printf "(COMMENT"
			for (i = 1; ++i <= NF; )
				printf " %s",$i
			printf ")\n"
			}
/^FontName /		{
			print "(COMMENT afm2pl conversion of " $2 ")"
			}
/^ItalicAngle /		{ if (! ItalicAngle) ItalicAngle = $2 }
/^XHeight /		{ if (! XHeight) XHeight = $2 }
/^IsFixedPitch /	{ 
			if ($2 == "true")
				IsFixedPitch = 1;
			}
/^SmallCapsFont/	{ SmallCapsFont = 1 }
/^MathFont/		{ MathFont = 1 }
/^NoKerns/		{ NoKerns = 1 }
/^NoLigs/		{ NoLigs = 1 }
/^ReEncode /		{ afm2tfm[$2] = $3; tfm2afm[$3] = $2 }
/^CharIC /		{ CharIC[$2] = $3 }

/^Ligature /		{
			kernsFor[$2] += 0
			n = ++ligsFor[$2]
			ligF[$2 "," n] = $3
			ligN[$2 "," n] = $4
			}
/^Kern /		{
			i = ++kernsFor[$2]
			kernF[$2 "," i] = $3
			kernA[$2 "," i] = $4
			}

/^C /			{
			# give unique id numbers to composite characters
			if ($2 == -1)
				$2 = --NextID;

			# if we don't have a re-encoding of this character
			# and its tfm slot isn't in use, use its id as
			# the tfm slot.
			if (! afm2tfm[$8] && ! tfm2afm[$2]) {
				afm2tfm[$8] = $2;
				tfm2afm[$2] = $8;
			}
			width[$8] = $5;
			id[$8] = $2;
			llx[$8] = $11;
			lly[$8] = $12;
			urx[$8] = $13;
			ury[$8] = $14;

			# handle any ligatures
			if (NF > 15 && !NoLigs && \
			    !SmallCapsFont || $8 !~ /^[a-z]$/) {
				for (i = 16; i <= NF; i++) {
					if ($i == "L") {
						kernsFor[$8] += 0
						n = ++ligsFor[$8]
						ligF[$8 "," n] = $++i
						ligN[$8 "," n] = $++i
					}
					while (i <= NF && $i != ";")
						++i
				}
			}
			}

/^KPX / && NoKerns == 0 {
		# Save kerning Information.
		if (!SmallCapsFont || ($2 !~ /^[a-z]$/ && $3 !~ /^[a-z]$/)) {
			i = ++kernsFor[$2]
			kernF[$2 "," i] = $3
			kernA[$2 "," i] = $4
		}
		}

END		{
		# Output the fontdimen parameters
		print "(FONTDIMEN"
		if (ItalicAngle != 0) {
			rad = (- ItalicAngle) / 57.2958
			slope = sin(rad)/cos(rad)
		} else
			slope = 0

		printf "    (SLANT R %g)\n", slope
		if (MathFont) {
			printf "    (SPACE R 0)\n"
			printf "    (STRETCH R 0)\n"
			printf "    (SHRINK R 0)\n"
		} else {
			printf "    (SPACE R %g)\n", width["space"]/1000
			printf "    (STRETCH R %g)\n", width["space"]/1500
			printf "    (SHRINK R %g)\n", width["space"]/2500
			printf "    (EXTRASPACE R %g)\n", width["space"]/2000
		}
		printf "    (XHEIGHT R %g)\n", XHeight/1000
		printf "    (QUAD R %g)\n", 1
		print "    )"

		# if this is a small-caps pseudo-font, make the
		# lower case letters have 80% of the width of
		# their upper case counterpart.
		if (SmallCapsFont) {
			for (i = 65; i <= 90; i++) {
				cu = tfm2afm[i]
				cl = tfm2afm[i+32]
				width[cl] = .8 * width[cu]
				llx[cl] = .8 * llx[cu]
				lly[cl] = .8 * lly[cu]
				urx[cl] = .8 * urx[cu]
				ury[cl] = .8 * ury[cu]
			}
		}
		# spit out kerns and ligatures
		print "(LIGTABLE"
		for (k in kernsFor) {
			c = afm2tfm[k]
			if ((64 < c && c < 91) || \
			    (96 < c && c < 123))
				cFmt = sprintf("C %c", c);
			else
				cFmt = sprintf("O %o", c)

			printf "    (LABEL %s)\n", cFmt

			n = ligsFor[k];
			for (i = 1; i <= n; i++) {
				f = ligF[k "," i]
				c = afm2tfm[f]
				if ((64 < c && c < 91) || \
				    (96 < c && c < 123))
					cFmt = sprintf("C %c", c);
				else
					cFmt = sprintf("O %o", c);

				printf "    (LIG %s", cFmt
				f = ligN[k "," i]
				c = afm2tfm[f]
				if ((64 < c && c < 91) || \
				    (96 < c && c < 123))
					printf(" C %c)\n", c);
				else
					printf(" O %o)\n", c);
			}
			n = kernsFor[k]
			for (i = 1; i <= n; i++) {
				f = kernF[k "," i]
				c = afm2tfm[f]
				if ((64 < c && c < 91) || \
				    (96 < c && c < 123))
					cFmt = sprintf("C %c", c);
				else
					cFmt = sprintf("O %o", c);

				printf "    (KRN %s R %g)\n", cFmt, \
					kernA[k "," i]/1000
			}
			print "    (STOP)"
		}
		print "    )"

		# spit out the character widths.
		for (i = 0; i < 128; i++) {
			if (width[tfm2afm[i]]) {
				if ((64 < i && i < 91) || \
				    (96 < i && i < 123))
					cFmt = sprintf("C %c", i);
				else
					cFmt = sprintf("O %o", i);

				ac = tfm2afm[i]
				printf "(CHARACTER %s\n", cFmt
				printf "    (CHARWD R %g)\n", width[ac]/1000
				printf "    (CHARHT R %g)\n", ury[ac]/1000

				# if the character had an explicit italic
				# correction, use it.  Otherwise,
				# if this is an italic font and the char
				# right edge is to the right of where the
				# font machinery would place the next
				# char, generate an italic correction for
				# the difference.

				if (CharIC[ac])
					printf "    (CHARIC R %g)\n", CharIC[ac]

				else if (ItalicAngle && urx[ac] > width[ac])
					printf "    (CHARIC R %g)\n",\
						(urx[ac]-width[ac])/1000

				# if the character extends below the baseline,
				# generate a depth item.
				if (lly[ac] < 0)
					printf "    (CHARDP R %g)\n", \
						-lly[ac]/1000
				
				print "    )"
			}
		}
		}
