#ifndef lint
static char rcsid[] =
    "@(#) $Header: dvi2ps.c,v 1.20 90/02/22 04:57:44 van Exp $ (LBL)";
#endif lint

/* 
 * AUTHOR(s) Mark Senn wrote the early versions of this program for the BBN
 * BitGraph.  Stephan Bechtolsheim, Bob Brown, Richard Furuta, James Schaad
 * and Robert Wells improved it.  Norm Hutchinson ported the program to the
 * Sun.  Neal Holtz ported it to the Apollo, and from there to producing
 * PostScript output. Scott Jones added intelligent font substitution. Howard
 * Trickey made it read GF files instead of PXL ones.  Van Jacobson adapted
 * the GF code to read PK files in addition to PXL ones, add support for 
 * Trevor Darrell's psfig & laserwriter font support (an enhanced version
 * of LA Carr's work on laserwriter fonts) and added user-selectable
 * font flushing intervals (to support long documents or complex figures).
 */

/*
 * Basic method: Using the local font cacheing machinery that was in the
 * previewer, we can easily manage to send the bitmap for each chracter only
 * once. Two passes are made over each page in the DVI file.  The first pass
 * simply outputs the bitmaps for all characters on that page that haven't
 * been sent before.  The second pass outputs all the character setting and
 * positioning commands.  This allows us to bracket the setting portion with
 * PostScript save's and restore's, thus reclaiming considerable virtual
 * memory after each page. 
 *
 * All coordinates are output in the PostScript system (TeX origin), and in
 * integer units of rasters (300/inch) -- except for character widths, which
 * are sent as floating point numbers. 
 *
 * About 2 pages of PostScript support code must be sent to the LaserWriter
 * before this stuff goes.  It is automatically included unless the -h option
 * is given. 
 */

typedef int BOOLEAN;

#define NEW(A) ((A *)calloc(1,sizeof(A)))

#ifdef DEBUG
int Debug = 0;
#endif

#define  DVIFORMAT        2
#define  TRUE             1
#define  FALSE            0
#define  FIRSTPXLCHAR     0
#define  LASTPXLCHAR      NPXLCHARS-1

#ifndef FONTAREA
#define FONTAREA         "/usr/local/fonts"
#endif
#ifndef FONTMAP
/* directory comes from FONTAREA or TEXFONTS or -a */
#define FONTMAP         "TeXPSfonts.map"
#endif

#define MAXBUILTINS     100

#ifndef HDRFILE
#define HDRFILE         "/usr/local/lib/ps/tex.ps"
#endif

#define  MAXOPEN         12	/* limit on number of open PXL files */

#define  PXLID         1001
#define  READ             4	/* for access() */
#define  RESOLUTION      300	/* default device resolution */
#define  STACKSIZE      100
#define  NONEXISTANT     -1	/* offset for PXL files not found */
#define  NO_FILE        (FILE *)-1
#define  NO_POS		-314159728
#define	 FLUSH_DEFAULT	1000	/* pages between font flushes */


#include "commands.h"
#include <sys/param.h>
#include <signal.h>
#include <stdio.h>
#include <ctype.h>
#include <sys/file.h>
#include "font.h"

int access ();
char *index ();
char *malloc ();
int free ();
char *rindex ();
char *sprintf ();
char *strcpy ();
char *strcat ();
char *logname ();

extern char *getenv();
extern int scanpk ();
extern void readpkbits ();

#define abs(a) ((a)<0?-(a):(a))
#define PixRound(x, conv) ((int) (((x) + ((conv) >> 1)) / (conv)))
#define RuleRound(x, conv) ((int)((x) + (conv) - 1) / (conv))

#define EQ(a,b) (strcmp(a,b)==0)

/* output a formatted string */
#define EMIT      (void) fprintf
/* output a simple string */
#define EMITS(s)  fputs(s,outfp)
/* output an escaped octal number */
#define EMITO(c)  PutOct(c)
/* output a decimal integer */
#define EMITN(n)  PutInt(n)
/* output a byte value in Hex */
#define EMITH(h)  putc(*(digit+((h>>4)&0xF)),outfp),\
                  putc(*(digit+(h&0xF)),outfp)
/* output a single character */
#define EMITC(c)  putc(c,outfp)
/* output a scaled X dimension */
#define EMITX(x)  PutInt(PixRound(x,hconv))
/* output a scaled Y dimension */
#define EMITY(y)  PutInt(PixRound(y,vconv))

/* formatted i/o was killing us, so build some tables */
char *digit = "0123456789ABCDEF";


/*
 * Note: Global procedures are declared here in alphabetical order, with
 * those which do not return values typed "void".  Their bodies occur in
 * alphabetical order following the main() procedure.  The names are kept
 * unique in the first 6 characters for portability. 
 */

void AbortRun ();
float ActualFactor ();
FILE *fopen ();
int ChkOpt ();			/* Check PostScript option for validity */
void CopyFile ();
void DecodeArgs ();
void DoSpecial ();
void EmitChar ();
void Fatal ();
void FindPostAmblePtr ();
void GetBytes ();
void GetFontDef ();
char *GetKeyStr ();
int GetKeyVal ();
int HasBeenRead ();
struct BuiltIn * IsBuiltIn ();
int IsSame ();
void lcase ();
void MoveDown ();
void MoveOver ();
int NoSignExtend ();		/* see cautionary note in code, re arithmetic
				 * vs logical shifts */
void OpenFontFile ();
FILE *OpenPath ();
char *PSname ();
void PutInt ();
int ReadFontDef ();
void ReadPostAmble ();
void SetChar ();
void SetFntNum ();
void SetPosn ();
void SetRule ();
void SetString ();
int SignExtend ();		/* see cautionary note in code, re arithmetic
				 * vs logical shifts */
void SkipFontDef ();
void Warning ();


int FirstPage = -1000000;	/* first page to print (uses count0) */
int LastPage = 1000000;		/* last page to print */

char filename[STRSIZE];		/* DVI file name */

int G_header = TRUE;		/* copy header file to output? */
char G_progname[STRSIZE];	/* program name */
int G_quiet = FALSE;		/* for quiet operation */
int G_nowarn = FALSE;		/* don't print out warnings */

int resolution = RESOLUTION;	/* postscript device 'resolution' in pixels */
int hconv, vconv;		/* converts DVI units to pixels */
int den;			/* denominator specified in preamble */
FILE *dvifp = NULL;		/* DVI file pointer */
int PreLoad = TRUE;		/* preload the font descriptions? */
struct font_entry *prevfont = NULL;	/* font_entry pointer, previous
					 * character */
struct font_entry *fontptr;	/* font_entry pointer */
struct font_entry *hfontptr = NULL;	/* font_entry pointer */
int h;				/* current horizontal position */
int hh;				/* current h in pixels */
int v;				/* current vertical position */
int vv;				/* current v in pixels */
int mag;			/* magnification specified in preamble */
char *Ifile[100];		/* files to include */
int ncopies = 1;		/* number of copies to print */
int ndone = 0;			/* number of pages converted */
int nif = 0;			/* number of files to include */
int nopen;			/* number of open PXL files */
int nps = 0;			/* number of PostScript commands to send */
int num;			/* numerator specified in preamble */
FILE *outfp = NULL;		/* output file */
struct font_entry *pfontptr = NULL;	/* previous font_entry pointer */
struct pixel_list pixel_files[MAXOPEN + 1];
/* list of open PXL file identifiers */
long postambleptr;		/* Pointer to the postamble */
FILE *pxlfp;			/* PXL file pointer */
char *PXLpath;			/* PXL path name for search */
char *TFMpath;			/* TFM path name for search */
long ppagep;			/* previous page pointer */
char *PScmd[100];		/* PostScript commands to send */
int Reverse = TRUE;		/* process DVI pages in reverse order ? */
char rootname[STRSIZE];		/* DVI filename without extension */

struct BuiltIn {
        struct BuiltIn *next;
	int Flush;		/* page where font def will be flushed */
        char TeXname[STRSIZE];	/* TeX Name of the LaserWriter font */
        char Postscript[STRSIZE];       /* Internal Name of the font */
	char Init[STRSIZE];	/* (optional) postscript init str */
} *BuiltIns;

int stdFmt = FALSE;		/* produce Adobe standard output */
int usermag = 0;		/* user specified magnification */
int flushint = FLUSH_DEFAULT;	/* pages between font flushes */
int flushpage = FLUSH_DEFAULT+1; /* next page to do font flush */
int pageNo;			/* page we're currently doing */



main (argc, argv)
	int argc;
	char *argv[];

{
	struct stack_entry {	/* stack entry */
		int h, v, w, x, y, z, hh, vv;	/* what's on stack */
	};


	int command;		/* current command */
	int count[10];		/* the 10 counters at begining of each page */
	long cpagep;		/* current page pointer */
	int Emitting = FALSE;	/* are we outputting typsetting instructions? */
	register int i;		/* command parameter; loop index */
	int k;			/* temporary parameter */
	char n[STRSIZE];	/* command parameter */
	int PassNo = 0;		/* which pass over the DVI page are we on? */
	int SkipMode = FALSE;	/* in skip mode flag */
	int sp;			/* stack pointer */
	struct stack_entry stack[STACKSIZE];	/* stack */
	char SpecialStr[STRSIZE];	/* "\special" strings */
	int val, val2;		/* temporarys to hold command information */
	int w;			/* current horizontal spacing */
	int x;			/* current horizontal spacing */
	int y;			/* current vertical spacing */
	int z;			/* current vertical spacing */

	nopen = 0;
	(void) strcpy (G_progname, argv[0]);

	if (! (TFMpath = getenv("TEXFONTS")))
		TFMpath = TFMAREA;

	DecodeArgs (argc, argv);
	ReadPSFontMap ();

	if ((i = NoSignExtend (dvifp, 1)) != PRE) {
		(void) fprintf (stderr, "\n");
		Fatal ("%s: PRE doesn't occur first--are you sure this is a DVI file?\n\n",
		       G_progname);
	}
	i = SignExtend (dvifp, 1);
	if (i != DVIFORMAT) {
		(void) fprintf (stderr, "\n");
		Fatal ("%s: DVI format = %d, can only process DVI format %d files\n\n",
		       G_progname, i, DVIFORMAT);
	}
	outfp = stdout;

	/* it is important that these be the very first things output !!! */

	if (stdFmt)
		EMIT (outfp, "%%!PS-Adobe-2.0\n");
	else
		EMIT (outfp, "%%!PS-\n");
	EMIT (outfp, "%%%%Title: %s\n", filename);
	EMIT (outfp, "%%%%Creator: %s\n", G_progname);
	EMIT (outfp, "%%%%Pages: (atend)\n");
	EMIT (outfp, "%%%%DocumentFonts: (atend)\n");
	EMIT (outfp, "%%%%EndComments\n");
	if (G_header)
		CopyFile (HDRFILE);

	for (i = 0; i < nif; i++)	/* copy all included files */
		CopyFile (Ifile[i]);

	EMIT (outfp, "%%%%EndProlog\n");

	EMIT (outfp, "TeXDict begin @start\n");

	for (i = 0; i < nps; i++)
		/* prefix valid PostScript options with a "@" */
		if (ChkOpt (PScmd[i]))
			EMIT (outfp, "@%s\n", PScmd[i]);
		else
			Fatal ("%s is an invalid option\n", PScmd[i]);

	if (ncopies > 1)
		EMIT (outfp, "%d @copies\n", ncopies);

	EMIT (outfp, "%d @startdoc\n", resolution);

	if (Reverse) {
		ReadPostAmble (PreLoad);
		(void) fseek (dvifp, ppagep, 0);
	} else {
		if (PreLoad) {
			ReadPostAmble (TRUE);
			(void) fseek (dvifp, (long) 14, 0);
		} else {
			num = NoSignExtend (dvifp, 4);
			den = NoSignExtend (dvifp, 4);
			mag = NoSignExtend (dvifp, 4);
			if (usermag > 0 && usermag != mag)
				(void) fprintf (stderr, "DVI magnification of %d over-ridden by user mag of %d\n", mag, usermag);
			if (usermag > 0)
				mag = usermag;
			hconv = DoConv (num, den, resolution);
			vconv = DoConv (num, den, resolution);
		}
		k = NoSignExtend (dvifp, 1);
		GetBytes (dvifp, n, k);
	}

	PassNo = 0;

	while (TRUE) {
		switch (command = NoSignExtend (dvifp, 1)) {

		case SET1:
		case SET2:
		case SET3:
		case SET4:
			val = NoSignExtend (dvifp, command - SET1 + 1);
			if (!SkipMode)
				SetChar (val, command, PassNo);
			break;

		case SET_RULE:
			val = NoSignExtend (dvifp, 4);
			val2 = NoSignExtend (dvifp, 4);
			if (Emitting)
				SetRule (val, val2, 1);
			break;

		case PUT1:
		case PUT2:
		case PUT3:
		case PUT4:
			val = NoSignExtend (dvifp, command - PUT1 + 1);
			if (!SkipMode)
				SetChar (val, command, PassNo);
			break;

		case PUT_RULE:
			val = NoSignExtend (dvifp, 4);
			val2 = NoSignExtend (dvifp, 4);
			if (Emitting)
				SetRule (val, val2, 0);
			break;

		case NOP:
			break;

		case BOP:
			cpagep = ftell (dvifp) - 1;
			for (i = 0; i <= 9; i++)
				count[i] = NoSignExtend (dvifp, 4);
			ppagep = NoSignExtend (dvifp, 4);

			h = v = w = x = y = z = 0;
			hh = vv = 0;
			sp = 0;
			fontptr = NULL;
			prevfont = NULL;

			if (count[0] < FirstPage || count[0] > LastPage)
				SkipMode = TRUE;
			else
				SkipMode = FALSE;

			Emitting = (PassNo != 0) && !SkipMode;

			if (!SkipMode) {
				if (PassNo == 0) {
					++pageNo;
					EMIT (outfp, "\n%%%%Page: %d %d\n",
					      count[0], pageNo);
					if (stdFmt)
						EMIT (outfp, "save\n");
					else
						EMIT (outfp, "%d @bop0\n",
						      count[0]);

					if (pageNo >= flushpage) {
						if (! stdFmt)
							EMIT (outfp,
							      "@flushfonts\n");
						flushpage += flushint;
					}
					if (!G_quiet)
						(void) fprintf (stderr, "[%d",
								count[0]);
				} else
					EMIT (outfp, "%d @bop1\n", count[0]);
			}
			break;

		case EOP:
			if (!SkipMode) {
				if (PassNo == 0) {	/* start second pass on
							 * current page */
					(void) fseek (dvifp, cpagep, 0);
					PassNo = 1;
				} else {	/* end of second pass, and of
						 * page processing */
					EMIT (outfp, "@eop\n");
					if (stdFmt)
						EMIT (outfp, "restore\n");
					if (!G_quiet) {
						(void) fprintf (stderr, "] ");
						if ((++ndone % 10) == 0)
							(void) fprintf (stderr, "\n");
					}
					PassNo = 0;
				}
			} else
				PassNo = 0;
			if (PassNo == 0 && Reverse) {
				if (ppagep > 0)
					(void) fseek (dvifp, ppagep, 0);
				else
					goto done;
			}
			break;

		case PUSH:
			if (sp >= STACKSIZE)
				Fatal ("stack overflow");
			stack[sp].h = h;
			stack[sp].v = v;
			stack[sp].w = w;
			stack[sp].x = x;
			stack[sp].y = y;
			stack[sp].z = z;
			sp++;
			break;

		case POP:
			--sp;
			if (sp < 0)
				Fatal ("stack underflow");
			h = stack[sp].h;
			v = stack[sp].v;
			w = stack[sp].w;
			x = stack[sp].x;
			y = stack[sp].y;
			z = stack[sp].z;
			break;

		case RIGHT1:
		case RIGHT2:
		case RIGHT3:
		case RIGHT4:
			val = SignExtend (dvifp, command - RIGHT1 + 1);
			if (Emitting)
				MoveOver (val);
			break;

		case W0:
			if (Emitting)
				MoveOver (w);
			break;

		case W1:
		case W2:
		case W3:
		case W4:
			w = SignExtend (dvifp, command - W1 + 1);
			if (Emitting)
				MoveOver (w);
			break;

		case X0:
			if (Emitting)
				MoveOver (x);
			break;

		case X1:
		case X2:
		case X3:
		case X4:
			x = SignExtend (dvifp, command - X1 + 1);
			if (Emitting)
				MoveOver (x);
			break;

		case DOWN1:
		case DOWN2:
		case DOWN3:
		case DOWN4:
			val = SignExtend (dvifp, command - DOWN1 + 1);
			if (Emitting)
				MoveDown (val);
			break;

		case Y0:
			if (Emitting)
				MoveDown (y);
			break;

		case Y1:
		case Y2:
		case Y3:
		case Y4:
			y = SignExtend (dvifp, command - Y1 + 1);
			if (Emitting)
				MoveDown (y);
			break;

		case Z0:
			if (Emitting)
				MoveDown (z);
			break;

		case Z1:
		case Z2:
		case Z3:
		case Z4:
			z = SignExtend (dvifp, command - Z1 + 1);
			if (Emitting)
				MoveDown (z);
			break;

		case FNT1:
		case FNT2:
		case FNT3:
		case FNT4:
			k = NoSignExtend (dvifp, command - FNT1 + 1);
			if (!SkipMode)
				SetFntNum(k);
			break;

		case XXX1:
		case XXX2:
		case XXX3:
		case XXX4:
			k = NoSignExtend (dvifp, command - XXX1 + 1);
			GetBytes (dvifp, SpecialStr, k);
			if (Emitting)
				DoSpecial (SpecialStr, k);
			break;

		case FNT_DEF1:
		case FNT_DEF2:
		case FNT_DEF3:
		case FNT_DEF4:
			k = NoSignExtend (dvifp, command - FNT_DEF1 + 1);
			if (PreLoad || HasBeenRead (k)) {
				SkipFontDef ();
			} else {
				ReadFontDef (k);
			}
			break;

		case PRE:
			Fatal ("PRE occurs within file");
			break;

		case POST:
			goto done;

		case POST_POST:
			Fatal ("POST_POST with no preceding POST");
			break;

		default:
			if (command >= FONT_00 && command <= FONT_63) {
				if (!SkipMode)
					SetFntNum (command - FONT_00);
			} else if (command >= SETC_000 && command <= SETC_127) {
				if (!SkipMode)
					SetString (command, PassNo);
			} else
				Fatal ("%d is an undefined command", command);
			break;

		}
	}
done:
	EMIT (outfp, "%%%%Trailer\n");
	EMIT (outfp, "@end\n");
	EMIT (outfp, "%%%%Pages: %d\n", pageNo);
	{
	register struct font_entry *fp;
	register char *cp = "%%DocumentFonts:";
	register struct BuiltIn *bi;

	for (fp = hfontptr; fp; fp = fp->next) {
		if (fp->font_type == PS && (bi = IsBuiltIn(fp->n)) &&
		    bi->Postscript[0] && ! bi->Init[0]) {
			EMIT(outfp, "%s %s\n", cp, bi->Postscript);
			cp = "%%+";
			bi->Postscript[0] = 0;
		}
	}
	if (!G_quiet)
		(void) fprintf (stderr, "\n");
	}
	EMIT (outfp, "%%%%EOF\n");
	exit (0);
	/*NOTREACHED*/

}


void
AbortRun (code)
	int code;
{
	exit (code);
}



/* compute the actual size factor given the approximation */
float
ActualFactor (unmodsize)
	int unmodsize;		/* actually factor * 1000 */
{
	float realsize;		/* the actual magnification factor */

	realsize = (float) unmodsize / 1000.0;
	/* a real hack to correct for rounding in some cases--rkf */
	if (unmodsize == 1095)
		realsize = 1.095445;	/* stephalf */
	else if (unmodsize == 1315)
		realsize = 1.314534;	/* stepihalf */
	else if (unmodsize == 1577)
		realsize = 1.577441;	/* stepiihalf */
	else if (unmodsize == 1893)
		realsize = 1.892929;	/* stepiiihalf */
	else if (unmodsize == 2074)
		realsize = 2.0736;	/* stepiv */
	else if (unmodsize == 2488)
		realsize = 2.48832;	/* stepv */
	else if (unmodsize == 2986)
		realsize = 2.985984;	/* stepiv */
	/*
	 * the remaining magnification steps are represented with sufficient
	 * accuracy already 
	 */
	return (realsize);
}

#define ISOPT(s) if( EQ(str,s) ) return( TRUE )

int
ChkOpt (str)
	char *str;
{
	/* lcase(str);  doesn't work */

	ISOPT ("note");		/* its a shame to build this into the program */
	ISOPT ("letter");
	ISOPT ("legal");
	ISOPT ("draft");
	ISOPT ("landscape");
	ISOPT ("manualfeed");
	return (FALSE);
}


/* copy a file straight through to output */

void
CopyFile (str)
	register char *str;
{
	register FILE *spfp;
	register int t;

	if ((spfp = fopen (str, "r")) == NULL) {
		(void) fprintf (stderr, "Unable to open file %s\n", str);
		return;
	}
	if (!G_quiet)
		(void) fprintf (stderr, " [%s", str);
	while ((t = getc (spfp)) != EOF) {
		EMITC (t);
	}
	(void) fclose (spfp);
	if (!G_quiet)
		(void) fprintf (stderr, "]");
}


void
DecodeArgs (argc, argv)
	int argc;
	char *argv[];
{
	int argind;		/* argument index for flags */
	char curarea[STRSIZE];	/* current file area */
	char curname[STRSIZE];	/* current file name */
	char *tcp, *tcp1;	/* temporary character pointers */
	int spooler = 0;	/* spooler */

	argind = 1;
	while (argind < argc) {
		tcp = argv[argind];
		if (*tcp == '-')
			switch (*++tcp) {

			case 'a':	/* a selects different pxl font area */
				PXLpath = argv[++argind];
				break;
#ifdef DEBUG
			case 'd':	/* d selects Debug output */
				Debug = TRUE;
				break;
#endif
			case 'f':	/* next arg is starting pagenumber */
				if (++argind >= argc || sscanf (argv[argind], "%d", &FirstPage) != 1)
					Fatal ("Argument is not a valid integer\n", 0);
				break;

			case 'h':	/* don't copy PostScript header file
					 * through to output */
				G_header = FALSE;
				break;

			case 'i':	/* next arg is a PostScript file to
					 * copy */
				if (++argind >= argc)
					Fatal ("No argument following -i\n", 0);
				Ifile[nif++] = argv[argind];
				break;

			case 'm':	/* specify magnification to use */
				switch (tolower (*++tcp)) {

				case '\0':	/* next arg is a
						 * magnification to use */
					if (++argind >= argc || sscanf (argv[argind], "%d", &usermag) != 1)
						Fatal ("Argument is not a valid integer\n", 0);
					break;
				case '0':
					usermag = 1000;
					break;
				case 'h':
					usermag = 1095;
					break;
				case '1':
					usermag = 1200;
					break;
				case '2':
					usermag = 1440;
					break;
				case '3':
					usermag = 1728;
					break;
				case '4':
					usermag = 2074;
					break;
				case '5':
					usermag = 2488;
					break;
				default:
					Fatal ("%c is a bad mag step\n", *tcp);
				}
				break;
			case 'n':	/* next arg is number of copies to
					 * print */
				if (++argind >= argc || sscanf (argv[argind], "%d", &ncopies) != 1)
					Fatal ("Argument is not a valid integer\n", 0);
				break;

			case 'o':	/* next arg is a PostScript command
					 * to send */
				if (++argind >= argc)
					Fatal ("No argument following -o\n", 0);
				PScmd[nps++] = argv[argind];
				break;

			case 'p':	/* p prohibits pre-font loading */
				PreLoad = 0;
				Reverse = FALSE;	/* must then process in
							 * forward order */
				break;

			case 'q':	/* quiet operation */
				G_quiet = TRUE;
				break;

			case 'r':	/* don't process pages in reverse
					 * order */
				Reverse = FALSE;
				break;
			case 's':	/* produce Adobe 'standard' output */
				stdFmt = TRUE;
				Reverse = FALSE;
				flushint = 1;
				flushpage = flushint+1;
				break;
			case 't':	/* next arg is ending pagenumber */
				if (++argind >= argc || sscanf (argv[argind], "%d", &LastPage) != 1)
					Fatal ("Argument is not a valid integer\n", 0);
				break;

			case 'w':	/* don't print out warnings */
				G_nowarn = TRUE;
				break;

			case 'F':	/* pages between font flushes */
				if (++argind >= argc ||
				    sscanf (argv[argind], "%d", &flushint) != 1)
					Fatal ("-F not followed by a number\n");
				if (flushint <= 0)
					flushint = FLUSH_DEFAULT;
				flushpage = flushint+1;
				break;

			case 'R':	/* printer 'resolution' */
				if (++argind >= argc ||
				    sscanf (argv[argind], "%d", &resolution) != 1)
					Fatal ("-R not followed by a number\n");
				if (resolution <= 0)
					resolution = RESOLUTION;
				break;

			case 'S':	/* we're an lpd filter */
				spooler = TRUE;
				break;

			default:
				(void) printf ("%c not a legal flag\n", *tcp);
			}

		else {
			/* split into directory + file name */
			tcp = rindex (argv[argind], '/');
			if (tcp == NULL) {
				curarea[0] = '\0';
				tcp = argv[argind];
			} else {
				(void) strcpy (curarea, argv[argind]);
				curarea[tcp - argv[argind] + 1] = '\0';
				tcp++;
			}

			/* split into file name + extension */
			(void) strcpy (curname, tcp);
			tcp1 = rindex (tcp, '.');
			if (tcp1 == NULL) {
				(void) strcpy (rootname, curname);
				(void) strcat (curname, ".dvi");
			} else {
				*tcp1 = '\0';
				(void) strcpy (rootname, curname);
				*tcp1 = '.';
			}

			(void) strcpy (filename, curarea);
			(void) strcat (filename, curname);

			if ((dvifp = fopen (filename, "r")) == NULL) {
				(void) fprintf (stderr, "\n");
				(void) fprintf (stderr, "%s: can't find DVI file \"%s\"\n\n", G_progname, filename);
				AbortRun (1);
			}
		}
		argind++;
	}

	if (dvifp == NULL) {
		register FILE *in = stdin;
		register FILE *out;
		register int c;

		if (! spooler) {
			(void) fprintf (stderr,
				 "\nusage: %s [-a area] [-c] [-h] [-o option] \
[-p] [-s] [-r] [-f n] [-t n] \
[-m0|h|1|2|3|4|  mag] \
[-a fontarea] dvifile\n",
				 G_progname);
			AbortRun (1);
		}
		/* create a temp file and copy stdin to it. */
		(void) strcpy (filename, "<stdin>");
		out = dvifp = tmpfile();

		while ((c = getc(in)) != EOF)
			putc (c, out);

		rewind (out);
	}
}



int 
DoConv (num, den, convResolution)
{
	register float conv;
	conv = ((float) num / (float) den) *
		((float) mag / 1000.0) *
		((float) convResolution / 254000.0);
	return ((int) (1.0 / conv + 0.5));
}



typedef enum {
	None, String, Integer, Number, Dimension
} ValTyp;

typedef struct {
	char *Key;		/* the keyword string */
	char *Val;		/* the value string */
	ValTyp vt;		/* the value type */
	union {			/* the decoded value */
		int i;
		float n;
	} v;
} KeyWord;

typedef struct {
	char *Entry;
	ValTyp Type;
} KeyDesc;

#define PSFILE 0
KeyDesc KeyTab[] = {{"psfile", String},
		    {"hsize", Dimension},
		    {"vsize", Dimension},
		    {"hoffset", Dimension},
		    {"voffset", Dimension},
		    {"hscale", Number},
		    {"vscale", Number}};

#define NKEYS (sizeof(KeyTab)/sizeof(KeyTab[0]))

void
DoSpecial (str, n)		/* interpret a \special command, made up of
				 * keyword=value pairs */
	char *str;
	int n;
{
	char spbuf[STRSIZE];
	char *sf = NULL;
	KeyWord k;
	int i;

	str[n] = '\0';
	spbuf[0] = '\0';

	SetPosn ();
	/* check for psfig "ps:" prefix */
	if (strncmp(str, "ps:", 3) == 0) {
		if (strncmp(str, "ps::[begin]", 11) == 0)
			EMIT (outfp, "%s\n", &str[11]);
		else if (strncmp(str, "ps::[end]", 9) == 0)
			EMIT (outfp, "%s\n", &str[9]);
		else if (strncmp(str, "ps: plotfile ", 13) == 0) {
			str += 13;
			for (sf = str; *sf && *sf != ' '; sf++) ;
			*sf = '\0';
			CopyFile (str);
		} else if (strncmp(str, "ps::", 4) == 0)
			EMIT (outfp, "%s\n", &str[4]);
		else
			EMIT (outfp, "%s\n", &str[3]);
		return;
	}
	EMITS ("@beginspecial\n");

	while ((str = GetKeyStr (str, &k)) != NULL) {	/* get all keyword-value
							 * pairs */
		/* for compatibility, single words are taken as file names */
		if (k.vt == None && access (k.Key, 0) == 0) {
			if (sf)
				Warning ("  More than one \\special file name given. %s ignored", sf);
			(void) strcpy (spbuf, k.Key);
			sf = spbuf;
		} else if (GetKeyVal (&k, KeyTab, NKEYS, &i) && i != -1) {
			if (i == PSFILE) {
				if (sf)
					Warning ("  More than one \\special file name given. %s ignored", sf);
				(void) strcpy (spbuf, k.Val);
				sf = spbuf;
			} else	/* the keywords are simply output as PS
				 * procedure calls */
				EMIT (outfp, "%f @%s\n", k.v.n, KeyTab[i].Entry);
		} else
			Warning ("  Invalid keyword or value in \\special - \"%s\" ignored", k.Key);
	}

	EMITS ("@setspecial\n");

	if (sf)
		CopyFile (sf);
	else
		Warning ("  No special file name provided.");

	EMITS ("@endspecial\n");
}


void
EmitChar (c, ce)		/* output a character bitmap */
	int c;
	struct char_entry *ce;
{
	register int i, j, cc;
	int nbpl, nwpl;
	register unsigned char *sl;
	double cw;

	/*
	 * Output in PostScript coord system (y +ive up, x +ive right) (0,0)
	 * of char bitmap at lower left.  Output scan lines from bottom to
	 * top 
	 */

	/* open font dict before first char */
	if (fontptr->Flush <= pageNo) {
		EMIT (outfp, "%g /%s @newfont\n", fontptr->scale,
		      fontptr->psname);
		fontptr->Flush = flushpage;
	}
	if (fontptr != prevfont) {	/* because this isn't done on pass 0 */
		EMIT (outfp, "%s @sf\n", fontptr->psname);
		prevfont = fontptr;
	}
	EMITS ("[<");
	cc = 2;

	nbpl = (ce->width + 7) >> 3;
	nwpl = (ce->width + 31) >> 5;
	for (i = ce->height; --i >= 0;) {
		sl = (unsigned char *) (ce->where.address.pixptr + i * nwpl);
		for (j = nbpl; --j >= 0;) {
			if (cc > 78) {
				EMITS ("\n  ");
				cc = 2;
			}
			EMITH (*sl);
			sl++;
			cc += 2;
		}
	}

	cw = (double)ce->tfmw / (double)hconv / fontptr->scale;
	EMIT (outfp, "> %d %d %d %d %.3f] %d @dc\n", nbpl << 3, ce->height,
		ce->xOffset, (((int) ce->height) - ce->yOffset) - 1, cw, c);

	ce->where.isloaded = flushpage;
	fontptr->ncdl++;

}



/*VARARGS1*/
void
Fatal (fmt, a, b, c)		/* issue a fatal error message */
	char *fmt;		/* format */
	char *a, *b, *c;	/* arguments */

{
	(void) fprintf (stderr, "\n");
	(void) fprintf (stderr, "%s: FATAL--", G_progname);
	(void) fprintf (stderr, fmt, a, b, c);
	(void) fprintf (stderr, "\n\n");
	AbortRun (1);
}



/*
 * this routine will move to the end of the file and find the start of the
 * postamble 
 */
void
FindPostAmblePtr (postambleptr)
	long *postambleptr;
{
	int i;

	(void) fseek (dvifp, (long) 0, 2);	/* goto end of file */
	*postambleptr = ftell (dvifp) - 4;
	(void) fseek (dvifp, *postambleptr, 0);

	while (TRUE) {
		(void) fseek (dvifp, --(*postambleptr), 0);
		if (((i = NoSignExtend (dvifp, 1)) != 223) &&
		    (i != DVIFORMAT))
			Fatal ("Bad end of DVI file");
		if (i == DVIFORMAT)
			break;
	}
	(void) fseek (dvifp, (*postambleptr) - 4, 0);
	(*postambleptr) = NoSignExtend (dvifp, 4);
	(void) fseek (dvifp, *postambleptr, 0);
}



void
GetBytes (fp, cp, n)		/* get n bytes from file fp */
	register FILE *fp;	/* file pointer */
	register char *cp;	/* character pointer */
	register int n;		/* number of bytes */

{
	while (n--)
		*cp++ = getc (fp);
}



/*
 * Read the font  definitions as they  are in the  postamble of the
 * DVI file.
 */
void
GetFontDef ()
{
	char *calloc ();
	unsigned char byte;

	while (((byte = NoSignExtend (dvifp, 1)) >= FNT_DEF1) &&
	       (byte <= FNT_DEF4)) {
		switch (byte) {
		case FNT_DEF1:
			ReadFontDef (NoSignExtend (dvifp, 1));
			break;
		case FNT_DEF2:
			ReadFontDef (NoSignExtend (dvifp, 2));
			break;
		case FNT_DEF3:
			ReadFontDef (NoSignExtend (dvifp, 3));
			break;
		case FNT_DEF4:
			ReadFontDef (NoSignExtend (dvifp, 4));
			break;
		default:
			Fatal ("Bad byte value in font defs");
			break;
		}
	}
	if (byte != POST_POST)
		Fatal ("POST_POST missing after fontdefs");
}


/*
 * extract first keyword-value pair from string (value part may be null)
 * return pointer to remainder of string return NULL if none found 
 */

char KeyStr[STRSIZE];
char ValStr[STRSIZE];

char *
GetKeyStr (str, kw)
	char *str;
	KeyWord *kw;
{
	char *s, *k, *v, t;

	if (!str)
		return (NULL);

	for (s = str; *s == ' '; s++);	/* skip over blanks */
	if (*s == '\0')
		return (NULL);

	for (k = KeyStr;	/* extract keyword portion */
	     *s != ' ' && *s != '\0' && *s != '=';
	     *k++ = *s++);
	*k = '\0';
	kw->Key = KeyStr;
	kw->Val = v = NULL;
	kw->vt = None;

	for (; *s == ' '; s++);	/* skip over blanks */
	if (*s != '=')		/* look for "=" */
		return (s);

	for (s++; *s == ' '; s++);	/* skip over blanks */
	if (*s == '\'' || *s == '\"')	/* get string delimiter */
		t = *s++;
	else
		t = ' ';
	for (v = ValStr;	/* copy value portion up to delim */
	     *s != t && *s != '\0';
	     *v++ = *s++);
	if (t != ' ' && *s == t)
		s++;
	*v = '\0';
	kw->Val = ValStr;
	kw->vt = String;

	return (s);
}


/*
 * get next keyword-value pair decode value according to table entry 
 */
int 
GetKeyVal (kw, tab, nt, tno)
	KeyWord *kw;
	KeyDesc tab[];
int nt;
int *tno;
{
	int i;
	char c = '\0';

	*tno = -1;

	for (i = 0; i < nt; i++)
		if (IsSame (kw->Key, tab[i].Entry)) {
			*tno = i;
			switch (tab[i].Type) {
			case None:
				if (kw->vt != None)
					return (FALSE);
				break;
			case String:
				if (kw->vt != String)
					return (FALSE);
				break;
			case Integer:
				if (kw->vt != String)
					return (FALSE);
				if (sscanf (kw->Val, "%d%c", &(kw->v.i), &c) != 1
				    || c != '\0')
					return (FALSE);
				break;
			case Number:
			case Dimension:
				if (kw->vt != String)
					return (FALSE);
				if (sscanf (kw->Val, "%f%c", &(kw->v.n), &c) != 1
				    || c != '\0')
					return (FALSE);
				break;
			}
			kw->vt = tab[i].Type;
			return (TRUE);
		}
	return (TRUE);
}



int
HasBeenRead (k)
	register int k;
{
	register struct font_entry *ptr;

	ptr = hfontptr;
	while (ptr && (ptr->k != k))
		ptr = ptr->next;

	return (ptr != NULL);
}



struct BuiltIn *
IsBuiltIn (a)
        char *a;
{
        struct BuiltIn *c;
        for (c = BuiltIns; c; c = c->next)
                if (strcmp (c->TeXname, a) == 0)
                        return (c);
        return (0);
}


int 
IsSame (a, b)			/* compare strings, ignore case */
	char *a, *b;
{
	for (; *a != '\0';)
		if (tolower (*a++) != tolower (*b++))
			return (FALSE);
	return (*a == *b ? TRUE : FALSE);
}



void
MoveDown (a)
	int a;
{
	v += a;
}


void
MoveOver (a)
	int a;
{
	h += a;
}



void
MoveRight (a)
	int a;
{
	h += a;
}



int
NoSignExtend (fp, n)		/* return n byte quantity from file fd */
	register FILE *fp;	/* file pointer */
	register int n;		/* number of bytes */

{
	register int x;		/* number being constructed */

	x = 0;
	while (n--) {
		x <<= 8;
		x |= getc (fp);
	}
	return (x);
}


/***********************************************************************
    The original version of this dvi driver reopened the font file  each
    time the font changed, resulting in an enormous number of relatively
    expensive file  openings.   This version  keeps  a cache  of  up  to
    MAXOPEN open files,  so that when  a font change  is made, the  file
    pointer, pxlfp, can  usually be  updated from the  cache.  When  the
    file is not found in  the cache, it must  be opened.  In this  case,
    the next empty slot  in the cache  is assigned, or  if the cache  is
    full, the least used font file is closed and its slot reassigned for
    the new file.  Identification of the least used file is based on the
    counts of the number  of times each file  has been "opened" by  this
    routine.  On return, the file pointer is always repositioned to  the
    beginning of the file.
  
***********************************************************************/
void
OpenFontFile ()
{
	register int i, least_used, current;

#ifdef DEBUG
	if (Debug)
		(void) printf ("Open Font file\n");
#endif
	if (pfontptr == fontptr)
		return;		/* we need not have been called */

	for (current = 1;
	     (current <= nopen) &&
	     (pixel_files[current].pixel_file_id != fontptr->font_file_id);
	     ++current);	/* try to find file in open list */

	if (current <= nopen) {
		pxlfp = pixel_files[current].pixel_file_id;
	} else {
		if (nopen < MAXOPEN)	/* just add it to list */
			current = ++nopen;
		else {		/* list full -- find least used file, close
				 * it, and reuse slot for new file */
			least_used = 1;
			for (i = 2; i <= MAXOPEN; ++i)
				if (pixel_files[least_used].use_count >
				    pixel_files[i].use_count)
					least_used = i;
			if (pixel_files[least_used].pixel_file_id != NO_FILE) {
				FILE *fid;
				struct font_entry *fp;

				fid = pixel_files[least_used].pixel_file_id;
				fp = hfontptr;	/* mark file as being closed
						 * in the entry */
				while (fp != NULL && fp->font_file_id != fid)
					fp = fp->next;
				if (fp == NULL)
					Fatal ("Open file %x not found in font entry list.\n", fid);
				else {
					fp->font_file_id = NULL;
				}
				(void) fclose (fid);
			}
			current = least_used;
		}
#ifdef	DEBUG
		if (Debug)
			(void) fprintf (stderr, "FONTFILE:%s\n", fontptr->name);
#endif
		if ((pxlfp = fopen (fontptr->name, "r")) == NULL) {
			Warning ("PXL file %s could not be opened",
				 fontptr->name);
			pxlfp = NO_FILE;
		} else {
		}
		pixel_files[current].pixel_file_id = pxlfp;
		pixel_files[current].use_count = 0;
	}
	pfontptr = fontptr;	/* make previous = current font */
	fontptr->font_file_id = pxlfp;	/* set file identifier */
	pixel_files[current].use_count++;	/* update reference count */
}


char *
PSname (a)
        char *a;
{
        struct BuiltIn *c;
        for (c = BuiltIns; c; c = c->next)
                if (strcmp (c->TeXname, a) == 0)
                        return (c->Postscript);
        return ("<NULLFONT>");
}


void
PutInt (n)			/* output an integer followed by a space */
	register int n;
{
	char buf[10];
	register char *b;

	if (n == 0)
		EMITC ('0');
	else {
		if (n < 0) {
			EMITC ('-');
			n = -n;
		}
		for (b = buf; n > 0;) {
			*b++ = digit[n % 10];
			n /= 10;
		}

		for (; b > buf;)
			EMITC (*--b);
	}

	EMITC (' ');
}


/*-->PutOct*/

void
PutOct (n)			/* output an 3 digit octal number preceded by
				 * a "\" */
	register int n;
{
	EMITC ('\\');
	EMITC (digit[(n & 0300) >> 6]);
	EMITC (digit[(n & 0070) >> 3]);
	EMITC (digit[n & 0007]);
}


int
ReadFontDef (k)
	int k;
{
	int t, i;
	register struct font_entry *tfontptr;
	register struct char_entry *tcharptr;

	if ((tfontptr = NEW (struct font_entry)) == NULL)
		Fatal ("can't malloc space for font_entry");
	tfontptr->next = hfontptr;
	fontptr = hfontptr = tfontptr;

	tfontptr->k = k;
	tfontptr->c = NoSignExtend (dvifp, 4);	/* checksum */
	tfontptr->s = NoSignExtend (dvifp, 4);	/* space size */
	tfontptr->d = NoSignExtend (dvifp, 4);	/* design size */
	tfontptr->a = NoSignExtend (dvifp, 1);	/* area length for font name */
	tfontptr->l = NoSignExtend (dvifp, 1);	/* device length */
#ifdef	DEBUG
	if (Debug) {
		(void) fprintf (stderr, "space size= %x\n", tfontptr->s);
		(void) fprintf (stderr, "design size= %x\n", tfontptr->d);
	}
#endif
	GetBytes (dvifp, tfontptr->n, tfontptr->a + tfontptr->l);
	tfontptr->n[tfontptr->a + tfontptr->l] = '\0';
	tfontptr->font_space = tfontptr->s / 6;	/* never used */
	/* compute magnification for font */
	tfontptr->font_mag = (int) ((ActualFactor ((int) (((double) tfontptr->s /
				     (double) tfontptr->d) * 1000.0 + 0.5)) *
					 ActualFactor (mag) *
					 (double) resolution) + 0.5);

        if (IsBuiltIn (tfontptr->n)) {
                FILE *tfmfl = NULL;
                char tfmname[1024];
		char work[1024];
                int lh, bc, ec, i, nw, tfmwidth[NPXLCHARS];
		double z = (double)tfontptr->s / (double)1048576.;
		double tw;

                (void) sprintf (tfontptr->psname, "%s.%d", tfontptr->n,
				 tfontptr->font_mag);
                /*** Read in TFM widths from TFM file (not PXL file) ***/
		(void) sprintf (work, "%s.tfm", tfontptr->n);
		tfmfl = OpenPath (TFMpath, work, tfmname);
		if (tfmfl == NULL) {
			tfontptr->font_file_id = NO_FILE;
			(void) fprintf (stderr,
				 "Can't find LaserWriter TFM file %s\n",
				 tfmname);
			return;
		}
                (void) fseek (tfmfl, 2L, 0);
                lh = NoSignExtend (tfmfl, 2);
                bc = NoSignExtend (tfmfl, 2);
                ec = NoSignExtend (tfmfl, 2);
                nw = NoSignExtend (tfmfl, 2);
                (void) fseek (tfmfl, (long) (24 + (lh + ec - bc + 1) * 4), 0);
                for (i = 0; i < nw; i++) {
			tw = (double)SignExtend(tfmfl,4);
                        tfmwidth[i] = (int)(tw * z);
		}
                (void) fseek (tfmfl, (long) lh * 4 + 24, 0);
                for (i = bc; i <= ec; i++) {
                        tfontptr->ch[i].tfmw = tfmwidth[NoSignExtend(tfmfl, 1)];
                        getc (tfmfl);
                        getc (tfmfl);
                        getc (tfmfl);
                }
		(void) fclose (tfmfl);
                tfontptr->font_type = PS;
                return;
        }

	if (!findfile (tfontptr, tfontptr->n))
		Fatal ("no font %s.%d", tfontptr->n, mag);

	if (tfontptr != pfontptr)
		OpenFontFile ();

	if (tfontptr->font_type == PXL) {	/* PXL */
		if (pxlfp == NO_FILE) {
			/* allow missing pxl files */
			tfontptr->magnification = 0;
			tfontptr->designsize = 0;
			for (i = FIRSTPXLCHAR; i <= LASTPXLCHAR; i++)
				tfontptr->ch[i].where.address.fileOffset =
					NONEXISTANT;
			return;
		}
		if ((t = NoSignExtend (pxlfp, 4)) != PXLID)
			Fatal ("PXL ID=%d, can only process PXL ID=%d files",
			       t, PXLID);
		(void) fseek (pxlfp, -20L, 2);
		t = NoSignExtend (pxlfp, 4);
		if ((tfontptr->c != 0) && (t != 0) && (tfontptr->c != t))
			Warning ("font = \"%s\",\n-->font checksum = %d,\n-->dvi checksum = %d",
				 tfontptr->name, tfontptr->c, t);
		tfontptr->magnification = NoSignExtend (pxlfp, 4);
		tfontptr->designsize = NoSignExtend (pxlfp, 4);

		(void) fseek (pxlfp, (long)(NoSignExtend (pxlfp, 4)*4), 0);

		for (i = FIRSTPXLCHAR; i <= LASTPXLCHAR; i++) {
			tcharptr = &(tfontptr->ch[i]);
			tcharptr->width = NoSignExtend (pxlfp, 2);
			tcharptr->height = NoSignExtend (pxlfp, 2);
			tcharptr->xOffset = SignExtend (pxlfp, 2);
			tcharptr->yOffset = SignExtend (pxlfp, 2);
			tcharptr->where.address.fileOffset = NoSignExtend (pxlfp, 4) * 4;
			tcharptr->tfmw = (double)NoSignExtend(pxlfp, 4)
					 * (double)tfontptr->s
					 / (double)(1 << 20);
		}
	} else {		/* PK */
		unsigned int checksum;

		for (i = FIRSTPXLCHAR; i <= LASTPXLCHAR; i++) {
			tfontptr->ch[i].where.address.fileOffset = NONEXISTANT;
		}
		if (pxlfp == NO_FILE)	/* allow missing pxl files */
			return;

		/* scan pk file and check char existance */
		checksum = scanpk(tfontptr, tfontptr->font_file_id);

		if ((tfontptr->c != 0) && (checksum != 0) && (tfontptr->c != checksum))
			Warning ("font = \"%s\",\n-->font checksum = %d,\n-->dvi checksum = %d",
				 tfontptr->name, tfontptr->c, checksum);
	}
}


ReadPSFontMap ()
{
	FILE *map;
	char TeXn[STRSIZE], PSn[STRSIZE], filen[STRSIZE], init[STRSIZE];
	struct BuiltIn *p = 0, *q;
	int c, status;

	map = OpenPath (TFMpath, FONTMAP, filen);
	if (map == NULL)
		Fatal ("Can't open TeX->PostScript font map %s\n", filen);
	c = 1;
	while ((status = fscanf(map, "%s %s", TeXn, PSn)) == 2) {
		q = (struct BuiltIn *) malloc (sizeof (struct BuiltIn));
		q->next = 0;
		(void) strcpy (q->TeXname, TeXn);
		(void) strcpy (q->Postscript, PSn);
		if (getc(map) != '\n') {
			status = fscanf(map, " \"%[^\n\"']\"", init);
			if (status == 1)
				(void) strcpy (q->Init, init);
		} else
			q->Init[0] = 0;

		if (p)
			p->next = q;
		else
			BuiltIns = q;
		p = q;
		c++;
	}
	if (status != EOF)
		Fatal ("Syntax error in %s, line %d\n", filen, c);
	(void) fclose (map);
}


void
ReadPostAmble (load)
	int load;
{
	FindPostAmblePtr (&postambleptr);
	if (NoSignExtend (dvifp, 1) != POST)
		Fatal ("POST missing at head of postamble");
#ifdef DEBUG
	if (Debug)
		(void) fprintf (stderr, "got POST command\n");
#endif
	ppagep = NoSignExtend (dvifp, 4);
	num = NoSignExtend (dvifp, 4);
	den = NoSignExtend (dvifp, 4);
	mag = NoSignExtend (dvifp, 4);
	if (usermag > 0 && usermag != mag)
		(void) fprintf (stderr, "DVI magnification of %d over-ridden by user mag of %d\n", mag, usermag);
	if (usermag > 0)
		mag = usermag;
	hconv = DoConv (num, den, resolution);
	vconv = DoConv (num, den, resolution);

	(void) NoSignExtend (dvifp, 4);/* height-plus-depth of tallest page */
	(void) NoSignExtend (dvifp, 4);/* width of widest page */
	if (NoSignExtend (dvifp, 2) >= STACKSIZE)
		Fatal ("Stack size is too small");
	(void) NoSignExtend (dvifp, 2);/* this reads the number of pages in */
	/* the DVI file */
#ifdef DEBUG
	if (Debug)
		(void) fprintf (stderr, "now reading font defs");
#endif
	if (load)
		GetFontDef ();
}


LoadAChar (c, ptr)
	int c;
	register struct char_entry *ptr;
{
	unsigned long *pr;
	register unsigned int nints;
#ifdef DEBUG
	int col, row, rowbytes;
#endif

	if (fontptr->font_type == PS) {
		if (fontptr->Flush <= pageNo) {
			register int i;
			struct BuiltIn *bi;

			/*
			 * if this font has a special init string
			 * (like a small caps font), output that
			 * string.  Then zero the init so we only
			 * do it once per font type (otherwise we'll
			 * try to do it once per size).
			 */
			if ((bi = IsBuiltIn(fontptr->n)) && bi->Init[0]
			     && bi->Flush <= pageNo) {
				EMIT (outfp, "/%s %s\n", bi->Postscript,
				      bi->Init);
				bi->Flush = flushpage;
			}
			/*
			 * Font Size is ActualFactor(mag) * s/d * Normal Size
			 * Normal Size is
			 *  d		(DVI UNITS)
			 *   * num/den	(10^-7 metres)
			 *   / 10^-5		(centimetres)
			 *   / 2.54		(inches)
			 *   * 300		DOTS
			 */
			EMIT (outfp, "/%s /%s ReEncodeForTeX /%s /%s %f TeXPSmakefont def\n",
			      PSname (fontptr->n),
			      fontptr->psname,
			      fontptr->psname,
			      fontptr->psname,
			      ActualFactor (mag) * fontptr->s * num / den
			      / 254000.0 * resolution);
			for (i = FIRSTPXLCHAR; i <= LASTPXLCHAR; i++)
				fontptr->ch[i].where.isloaded = flushpage;
			fontptr->ncdl++;
			fontptr->Flush = flushpage;
		}
		return;
	}

	if (ptr->where.isloaded && ptr->where.isloaded <= pageNo) {
		EmitChar (c, ptr);
		return;
	}
	if (ptr->where.address.fileOffset == NONEXISTANT) {
		ptr->where.address.pixptr = NULL;
		ptr->where.isloaded = 999999;
		return;
	}
	OpenFontFile ();
	(void) fseek (pxlfp, ptr->where.address.fileOffset, 0);

	if (fontptr->font_type == PXL) {
		nints = ((ptr->width + 31) >> 5) * ptr->height;
		pr = (unsigned long *) malloc (nints * sizeof (*pr));
		if (pr == NULL)
			Fatal ("Unable to allocate memory for char\n");
		ptr->where.address.pixptr = pr;
		(void) fread ((char *)pr, 4, (int)nints, pxlfp);
	} else
		readpkbits (pxlfp, ptr);

	EmitChar (c, ptr);
}

void
SetChar (c, command, PassNo)
	int c, command, PassNo;
{
	register struct char_entry *ptr;

	ptr = &(fontptr->ch[c]);
	if (ptr->where.isloaded <= pageNo)
		LoadAChar (c, ptr);
	if (PassNo == 0)
		return;

	SetPosn ();
	if (fontptr->font_file_id != NO_FILE) {	/* ignore missing fonts */
		if (fontptr != prevfont) {
			EMIT (outfp, "%s @sf\n", fontptr->psname);
			prevfont = fontptr;
		}
		EMITN (c);
		EMITS ("c\n");
	}
	if (command <= SET4) {
		hh += ptr->tfmw;
		MoveRight (ptr->tfmw);
	}
}


/*
 * this routine is used to specify the font to be used in printing future
 * characters 
 */
void
SetFntNum (k)
	register int k;
{
	fontptr = hfontptr;
	while (fontptr && (fontptr->k != k))
		fontptr = fontptr->next;
	if (fontptr == NULL)
		Fatal ("font %d undefined", k);
}

/*
 * Make the current pixel position agree with TeX's position.
 */
void
SetPosn ()
{
	register int xpos, ypos;

	xpos = PixRound(h, hconv);
	ypos = PixRound(v, vconv);
	if ((ypos * vconv) == vv) {
		/* use relative movement if just moving horizontally */
		if (xpos = PixRound (h - hh, hconv)) {
			EMIT (outfp, "%d r ", xpos);
			hh += xpos * hconv;
		}
	} else {
		EMIT (outfp, "%d %d p ", xpos, ypos);
		vv = ypos * vconv;
		hh = xpos * hconv;
	}
}


/* draw a rule */

void
SetRule (a, b, Set)
	int a, b;
	BOOLEAN Set;

{
	int q;		/* b, rounded to pixel */

	if (a > 0 && b > 0) {
		SetPosn ();	/* lower left corner */
		q = RuleRound (b, hconv);
		EMITN (q);	/* width */
		EMITN (RuleRound (a, vconv));	/* height */
		EMITS ("ru\n");
	}
	if (Set) {
		hh += q;
		MoveRight (b);
	}
}



void
SetString (firstch, PassNo)	/* read and set a consecutive string of chars */
	int firstch, PassNo;
{
	char s[256];
	register char *sp;
	register int c;
	register struct char_entry *ptr;
	register int len;

	/* read entire string of chars */

	for (c = firstch, sp = s;
	     sp < &s[256] && c >= SETC_000 && c <= SETC_127;) {
		*sp++ = c;
		/* ensure that character is loaded */
		if (fontptr->ch[c].where.isloaded <= pageNo)
			LoadAChar (c, &fontptr->ch[c]);
		c = NoSignExtend (dvifp, 1);
	}
	(void) ungetc (c, dvifp);
	len = sp - s;		/* NULL's are valid chars, so cant use for
				 * string termination */
	if (PassNo == 0)
		return;
	if (fontptr->font_file_id == NO_FILE)
		return;
	if (fontptr != prevfont) {
		EMIT (outfp, "%s @sf ", fontptr->psname);
		prevfont = fontptr;
	}
	for (sp = s; sp < s + len; sp++) {
		if (sp == s + len - 1) {
			SetChar (*sp, SET1, PassNo);
		} else {
			register char *cmd = ")s\n";
			register int xpos, ypos;

			xpos = PixRound(h, hconv);
			ypos = PixRound(v, vconv);
			if ((ypos * vconv) != vv)
				SetPosn ();
			else if (xpos = PixRound (h - hh, hconv)) {
				EMIT (outfp, "%d", xpos);
				hh += xpos * hconv;
				cmd = ")m\n";
			}
			EMITC ('(');
			while (sp < s + len) {
				c = *sp++;
				ptr = &(fontptr->ch[c]);
				h += ptr->tfmw;
				hh += ptr->tfmw;
				if (c < ' ' || c >= 0177)
					EMITO (c);
				else if (c == '(' || c == ')' || c == '\\') {
					EMITC ('\\');
					EMITC (c);
				} else
					EMITC (c);
			}
			EMITS (cmd);
		}
	}

}


int
SignExtend (fp, n)		/* return n byte quantity from file fd */
	register FILE *fp;	/* file pointer */
	register int n;		/* number of bytes */

{
	register int x;		/* number being constructed */

	x = (unsigned)getc(fp);	/* get first (high-order) byte */
	switch (n) {
	case 1:
		return (x < 128? x : 256 - x);
	case 2:
		if (x >= 128)
			x |= (65535 << 8);
		x = (x << 8) | (unsigned) getc(fp);
		break;
	case 3:
		if (x >= 128)
			x |= (255 << 8);
		x = (x << 8) | (unsigned) getc(fp);
		x = (x << 8) | (unsigned) getc(fp);
		break;
	case 4:
		x = (x << 8) | (unsigned) getc(fp);
		x = (x << 8) | (unsigned) getc(fp);
		x = (x << 8) | (unsigned) getc(fp);
		break;
	}
	return (x);
}



void
SkipFontDef ()
{
	int a, l;
	char n[STRSIZE];

	(void) NoSignExtend (dvifp, 4);
	(void) NoSignExtend (dvifp, 4);
	(void) NoSignExtend (dvifp, 4);
	a = NoSignExtend (dvifp, 1);
	l = NoSignExtend (dvifp, 1);
	GetBytes (dvifp, n, a + l);
}


/*VARARGS1*/
void
Warning (fmt, a, b, c)		/* issue a warning */
	char *fmt;		/* format */
	char *a, *b, *c;	/* arguments */
{
	if (! G_nowarn) {
		(void) fprintf (stderr, fmt, a, b, c);
		(void) fprintf (stderr, "\n");
	}
}


FILE *
OpenPath (path, basename, foundname)
	char *path;
	char *basename;
	char *foundname;
{
	char work[1024];
	char *cp;
	FILE *fp;

	if (path) {
		(void) strcpy (work, path);
		path = work;
	}
	while (path && *path) {
		if (cp = index (path, ':'))
			*cp++ = 0;

		(void) sprintf (foundname, "%s/%s", path, basename);
		if (fp = fopen (foundname, "r"))
			return (fp);

		path = cp;
	}
	(void) strcpy (foundname, basename);
	return NULL;
}
