/*
 * findfile - try to find the raster image for some font at some mag.
 */

#ifndef lint
static char rcsid[] =
"@(#) $Header: findfile.c,v 1.4 88/08/30 04:48:22 van Exp $ (LBL)";
#endif

#include <stdio.h>
#include <sys/types.h>
#include <sys/dir.h>
#include "font.h"
#include <sys/file.h>
#include <strings.h>

#define	 MAXPATHLEN	1024

extern char *PXLpath;



struct fontSize {
	int mag;		/* this size */
	struct fontSize *next;	/* next larger size */
	char *filename;		/* full file name of this size */
};

struct fontNames {
	char *name;		/* base name of font (e.g., cmr10) */
	struct fontSize size;	/* list of pixel file magnifications */
};
	
extern char *calloc();
extern char *malloc();

static char *nmcopy ();
static int hash ();
static struct fontSize *findFont ();
static void add_fonts_on_path ();
static void add_fonts_in_directory ();
static void add_font ();
static void print_font_table ();


/* true if it found font file, false otherwise; name found returned in fp */

int 
findfile (fp, name)
	struct font_entry *fp;
	char *name;
{
	extern char *getenv ();
	char *cp;
	struct fontSize *sp;
	static int initialized;

	/*
	 * if we've got a full pathname, just check that it exists.
	 */
	if (*name == '/' || *name == '.' &&
	    (! strncmp (name, "./", 2) || ! strncmp (name, "../", 3))) {
		(void) strcpy (fp->name, name);
		return (access (name, R_OK) == 0);
	}
	/*
	 * not a full pathname -- search for the file.
	 */
	if (! initialized) {
		/* build the pixel file database */
		add_fonts_on_path (PXLpath);
		if (cp = getenv ("PXLPATH"))
			add_fonts_on_path (cp);
		else
			add_fonts_on_path (FONTAREA);

#ifdef notdef
		print_font_table ();
#endif notdef
		initialized = 1;
	}
	if (!(sp = findFont (name, fp->font_mag)))
		return (0);

	(void) strcpy (fp->name, sp->filename);
	(void) sprintf (fp->psname, "%s.%d", fp->n, fp->font_mag);
	/*
	 * the font scalefactor is the ratio of the size we
	 * wanted to the size we got (which will be 1.0 if we
	 * found a pixel file of the right size).
	 */
	fp->scale = ((double)fp->font_mag) / ((double)sp->mag);
	return (1);
}


/* make a copy of a string in dynamically allocated memory */
static char *
nmcopy (str)
	register char *str;
{
	return (strcpy (malloc (strlen (str) + 1), str));
}


#define MODULUS 16381	/* Must be prime! */

static struct fontNames nameTable[MODULUS];

/* "hashpjw" from Aho, Sethi & Ullman, p.436 */
static int
hash (str)
	register unsigned char *str;
{
	register unsigned int h = 0, g;

	while (g = *str++) {
		h = (h << 4) + g;
		if (g = h & 0xf0000000) {
			h = h ^ (g >> 24);
			h = h ^ g;
		}
	}
	return (h % MODULUS);
}

static void
add_font (name, mag, filename)
	char *name;
	int mag;
	char *filename;
{
	register int h = hash (name);
	register struct fontSize *sp;
	struct fontSize *sp2;

	while (nameTable[h].name && strcmp(nameTable[h].name, name)) {
		h = ++h % MODULUS;
	}
	if (! nameTable[h].name) {
		/* didn't find name in table - add it */
		nameTable[h].name = nmcopy (name);
		nameTable[h].size.mag = mag;
		nameTable[h].size.filename = nmcopy (filename);
		return;
	}
	/* found entry - find right place to insert size */
	for (sp = &nameTable[h].size;
	     sp->next && sp->next->mag <= mag;
	     sp = sp->next)
		;
	if (sp->mag == mag)
		return;		/* we've already got this size */

	sp2 = (struct fontSize *) malloc (sizeof(struct fontSize));
	if (sp->mag < mag) {
		/*
		 * if there's another entry in the list, it must be
		 * larger than us.  add the new entry after the current entry
		 */
		sp2->next = sp->next;
		sp2->mag = mag;
		sp2->filename = nmcopy (filename);
	}
	else {
		/* add the new entry in front of the current entry */
		*sp2 = *sp;
		sp->mag = mag;
		sp->filename = nmcopy (filename);
	}
	sp->next = sp2;
}

static struct fontSize *
findFont (name, mag)
	char *name;
	int mag;
{
	register int h = hash (name);
	register struct fontSize *sp;
	int i;

	while (nameTable[h].name && strcmp(nameTable[h].name, name)) {
		h = ++h % MODULUS;
	}
	if (! nameTable[h].name) {
		/* didn't find name in table */
		fprintf (stderr, "-couldn't find %s at mag %d\n", name, mag);
		return (0);
	}
	/* found name - look for right size */
	for (sp = &nameTable[h].size;
	     sp->next && sp->next->mag <= mag;
	     sp = sp->next)
		;
	if (sp->mag != mag) {
		i = mag - sp->mag;
		if (sp->next && i > (sp->next->mag - mag)) {
			sp = sp->next;
			i = sp->mag - mag;
		}
		/*
		 * if there's more than 1% size error, announce
		 * the substitution.
		 */
		if (i < 0)
			i = -i;
		if ((i*100 - 1)/mag > 0)
			fprintf (stderr,
				 "-substituting %s at mag %d for mag %d\n",
				 name, sp->mag, mag);
	}
	return (sp);
}


/*
 * add all the pixel and pk fonts found in directories on the
 * search path "path".  In addition to the contents of the path
 * directories, "pk" and "pixel" subdirectories are searched for
 * font files.
 */
static void
add_fonts_on_path (path)
	char *path;
{
	char *cp;
	char work[MAXPATHLEN];
	
	while (path && *path) {
		if (cp = index (path, ':'))
			*cp++ = 0;
		
		add_fonts_in_directory (path);
		(void) sprintf (work, "%s/pk", path);
		add_fonts_in_directory (work);
		(void) sprintf (work, "%s/pixel", path);
		add_fonts_in_directory (work);

		path = cp;
	}
}


static void
add_fonts_in_directory (dname)
	char *dname;
{
	DIR *dir;
	struct direct *dp;
	char fullname[MAXPATHLEN];
	char *cp;
	int mag;

	if (! (dir = opendir (dname)))
		return;

	while (dp = readdir (dir)) {
		if (strcmp (dp->d_name + dp->d_namlen - 2, "pk") == 0) {
			/* we've got a pk file */
			if (! (cp = rindex (dp->d_name, '.')))
				continue;
			sprintf (fullname, "%s/%s", dname, dp->d_name);
			mag = atoi (cp + 1);
			*cp = '\0';
			add_font (dp->d_name, mag, fullname);
		}
		else if (strcmp (dp->d_name + dp->d_namlen - 3, "pxl") == 0) {
			/* we've got a pxl file */
			if (! (cp = rindex (dp->d_name, '.')))
				continue;
			sprintf (fullname, "%s/%s", dname, dp->d_name);
			mag = (atoi (cp + 1) + 2) / 5;
			*cp = '\0';
			add_font (dp->d_name, mag, fullname);
		}
	}
	closedir (dir);
}

static void
print_font_table ()
{
	struct fontSize *sp;
	int i;

	for (i = 0; i < MODULUS; ++i)
		if (nameTable[i].name) {
			printf ("%s\n", nameTable[i].name);
			for (sp = &nameTable[i].size; sp; sp = sp->next)
				printf ("%6d\t%s\n", sp->mag, sp->filename);
		}
}
