#ifndef lint
static char rcsid[] =
    "@(#) $Header: pk.c,v 1.7 87/10/28 17:37:48 leres Exp $ (LBL)";
#endif lint
/*
 * This file contains routines for reading character bitmaps from PK
 * files.
 */

#include <sys/types.h>
#include <stdio.h>
#include "defs.h"
#include "font.h"

#define _READ		(getc(pkin))

#define UNSIGNED_1	_READ
#define SIGNED_1	((_READ << 24) >> 24)

#define UNSIGNED_2(tmp)	(tmp = _READ, (tmp << 8) | _READ)
#define SIGNED_2(tmp)	((UNSIGNED_2(tmp) << 16) >> 16)

#define UNSIGNED_3(tmp)	(tmp = UNSIGNED_2(tmp), (tmp << 8) | _READ)
#define SIGNED_3(tmp)	((UNSIGNED_3(tmp) << 8) >> 8)

/* since we have 4 byte int's, no sign extension is required... */
#define UNSIGNED_4(tmp)	(tmp = UNSIGNED_3(tmp), (tmp << 8) | _READ)
#define SIGNED_4(tmp)	UNSIGNED_4(tmp)


/*
 * the macros that read pk values need these variables.
 */
static u_short	_nyb_buf,
		_nyb_flag,
		_pk_repeat;


/*
 * get the next nybble of the file.  This macro requires
 * that 2 integers named _nyb_buf and _nyb_flag be allocated elsewhere in
 * the program.  In addition, _nyb_flag must be initialized to 0.
 */
#define GET_NYB ((_nyb_flag ^= 1) ? \
		    ((_nyb_buf = (unsigned)getc(pkin)) >> 4) : \
		    (_nyb_buf & 0xf))

/*
 * The quantity to be packed into nybbles may require an odd number of
 * nybbles which will cause the nybble fetching macro to get out of
 * sync.  The following macro ``clears'' the state of the nybble fetching
 * routine and should be executed whenever transitioning from nybble to
 * byte (or other) quantities.
 */
#define CLEAR_NYB	_nyb_flag = 0;

/*
 * this macro gets a PK ``packed number'' from fp and puts it into x.
 * It is an adaption of an algorithm presented in Tugboat V. 6, No. 3.
 */
#define GET_PACKED(x) x = get_packed(pkin, dyn_f)

static u_int
get_packed(pkin, dyn_f)
	register FILE *pkin;
	register u_int dyn_f;
{
	register int i, j;

	i = GET_NYB;
	if (i == 0) {
		/*	
		 * we have an arbitrarily long number.  scan to
		 * find the first non-zero nybble which is the
		 * count of the nybbles in this value.
		 */
		do {
			j = GET_NYB;
			i++;
		} while (j == 0);

		while (--i >= 0)
			j = (j << 4) + GET_NYB;

		return (j + 193 - 15 * dyn_f);
	} else if (i <= dyn_f) {
		/* this nybble is the number we want. */
		return (i);
	} else if (i < 14) {
		return (i + 15 * (i - dyn_f - 1) + GET_NYB);
	} else if (i == 14) {
		_pk_repeat = get_packed (pkin, dyn_f);
	} else {
		_pk_repeat = 1;
	}
	return (i);
}


int 
scanpk (fp, pkin)
	register struct font_entry *fp;
	register FILE *pkin;
{
	register unsigned c, cc;
	long pl, addr;
	register u_int i;
	u_int hppp, vppp;
	int checksum;
	u_int tfmw;
	double cscale;

	fseek (pkin, 0L, 0);	/* make sure at the beginning of pk file */
	if (UNSIGNED_1 != PK_PRE) {
		fprintf (stderr, "dvi2ps: pk font file %s doesn't start with PRE\n",
			 fp->name);
		exit (1);
	}
	if (UNSIGNED_1 != PK_ID) {
		fprintf (stderr, "dvi2ps: pk font file %s wrong version\n",
			 fp->name);
		exit (1);
	}
	fseek (pkin, (long) UNSIGNED_1, 1);	/* skip comment */
	fp->designsize = UNSIGNED_4(i);	/* ds[4] */
	checksum = UNSIGNED_4(i);	/* checksum[4] */
	hppp = UNSIGNED_4(i);		/* hppp[4] */
	vppp = UNSIGNED_4(i);		/* vppp[4] */
	cscale = (double) fp->s / (double)(1 << 20);

	while ((c = UNSIGNED_1) != EOF) {
		if (c >= PK_XXX1) {	/* commands are just skipped */
			switch (c) {
			case PK_XXX1:	/* pk_xxx1 k[1] x[k] */
				fseek (pkin, (long) UNSIGNED_1, 1);
				break;
			case PK_XXX2:	/* pk_xxx2 k[2] x[k] */
				fseek (pkin, (long) UNSIGNED_2(i), 1);
				break;
			case PK_XXX3:	/* pk_xxx3 k[3] x[k] */
				fseek (pkin, (long)UNSIGNED_3(i), 1);
				break;
			case PK_XXX4:	/* pk_xxx4 k[4] x[4] */
				fseek (pkin, (long)SIGNED_4(i), 1);
				break;
			case PK_YYY:	/* pk_yyy y[4] */
				(void) UNSIGNED_4(i);
				break;
			case PK_POST:
				return (checksum);
			case PK_PRE:
				fprintf (stderr,
					 "dvi2ps: pk font file %s has extra PRE\n",
					 fp->name);
				exit (1);
			default: /* do nothing */ ;
			}
		} else {	/* flag byte */
			switch (c & 0x07) {	/* check flag byte */
			case 0:/* short form */
			case 1:
			case 2:
			case 3:
				/* length */
				pl = (long) UNSIGNED_1
				     + (long) ((c & 0x03) << 8);
				cc = UNSIGNED_1;	/* char. code */
				tfmw = UNSIGNED_3(i); /* tfm width */
				(void) UNSIGNED_1;	 /* x-escapement */
				fp->ch[cc].width = UNSIGNED_1;
				fp->ch[cc].height = UNSIGNED_1;
				fp->ch[cc].xOffset = SIGNED_1;
				fp->ch[cc].yOffset = SIGNED_1;
				addr = ftell (pkin);
				pl -= 8;
				break;

			case 4:/* extended short form */
			case 5:
			case 6:
				pl = (long) UNSIGNED_2(i)
				     + (long) ((c & 0x03) << 16);
				cc = UNSIGNED_1;	/* char. code */
				tfmw = UNSIGNED_3(i); /* tfm width */
				(void) UNSIGNED_2(i);	 /* x-escapement */
				fp->ch[cc].width = UNSIGNED_2(i);
				fp->ch[cc].height = UNSIGNED_2(i);
				fp->ch[cc].xOffset = SIGNED_2(i);
				fp->ch[cc].yOffset = SIGNED_2(i);
				addr = ftell (pkin);
				pl -= 13;
				break;

			case 7:/* long form */
				pl = UNSIGNED_4(i);
				cc = UNSIGNED_4(i);	/* char. code */
				tfmw = UNSIGNED_4(i);
				(void) UNSIGNED_4(i);	 /* x-escapement */
				fp->ch[cc].width = UNSIGNED_4(i);
				fp->ch[cc].height = UNSIGNED_4(i);
				fp->ch[cc].xOffset = SIGNED_4(i);
				fp->ch[cc].yOffset = SIGNED_4(i);
				addr = ftell (pkin);
				pl -= 24;
				break;
			}
			fp->ch[cc].tfmw = (int)(((double)tfmw * cscale) + 0.5);
			fp->ch[cc].where.isloaded = 0;
			fp->ch[cc].where.flags = c;
			fp->ch[cc].where.address.fileOffset = addr;;
			fseek (pkin, pl, 1);	/* skip until next flag byte */
		}
	}
	return (checksum);
}

#define ROWADDR(pos) &sp[((pos) / cw) * rowsize]

#define FILL(row_pos, nbits) if (nbits) \
				fill(ROWADDR(row_pos), (row_pos) % cw, (nbits))
#define FILLR(row_pos, nrows) if (nrows) \
				fillw(ROWADDR(row_pos), (nrows), cw)

#define COPY(rowp, bp, rc) rowcopy(ROWADDR(rowp), ROWADDR(bp), rc, rowsize)

static u_int mask[] = {
#ifdef vax
	0xffffffff, 0xffffff7f, 0xffffff3f, 0xffffff1f,
	0xffffff0f, 0xffffff07, 0xffffff03, 0xffffff01,
	0xffffff00, 0xffff7f00, 0xffff3f00, 0xffff1f00,
	0xffff0f00, 0xffff0700, 0xffff0300, 0xffff0100,
	0xffff0000, 0xff7f0000, 0xff3f0000, 0xff1f0000,
	0xff0f0000, 0xff070000, 0xff030000, 0xff010000,
	0xff000000, 0x7f000000, 0x3f000000, 0x1f000000,
	0x0f000000, 0x07000000, 0x03000000, 0x01000000
#else
	0xffffffff, 0x7fffffff, 0x3fffffff, 0x1fffffff,
	0x0fffffff, 0x07ffffff, 0x03ffffff, 0x01ffffff,
	0x00ffffff, 0x007fffff, 0x003fffff, 0x001fffff,
	0x000fffff, 0x0007ffff, 0x0003ffff, 0x0001ffff,
	0x0000ffff, 0x00007fff, 0x00003fff, 0x00001fff,
	0x00000fff, 0x000007ff, 0x000003ff, 0x000001ff,
	0x000000ff, 0x0000007f, 0x0000003f, 0x0000001f,
	0x0000000f, 0x00000007, 0x00000003, 0x00000001,
#endif
};

/*
 * fill "nbits" bits of an array "sp" starting at bit position bit_pos.
 */
static void
fill (sp, bit_pos, nbits)
	register u_long *sp;
	register int bit_pos;
	register int nbits;
{
	register int i, j;

	sp += bit_pos / 32;
	i = bit_pos % 32;
	j = (nbits + bit_pos) % 32;
	if (nbits <= 32) {
		if (i < j)
			*sp |= mask[i] & ~mask[j];
		else if (j == 0)
			*sp |= mask[i];
		else {
			*sp++ |= mask[i];
			*sp = ~mask[j];
		}
	} else {
		*sp++ |= mask[i];
		/*
		 * there's some algebra involved in this next line.
		 * we want to figure out how many full words to fill.
		 * we did 32 - i bits in the first word and will do j
		 * bits in the last word so the remainder is
		 * nbits - (32 - i) - j bits or
		 * (nbits + i - j - 32) / 32 words
		 * This "for" loop should be initialized with the
		 * preceding but we save one subtract by moving
		 * the -1 into the test condition.
		 */
		for (i = (nbits + i - j) / 32; --i > 0; )
			*sp++ = ~0;
		if (j)
			*sp = ~mask[j];
	}
}

/*
 * fill "nrows" starting at sp with a field of one bits "nbits" wide.
 */
static void
fillw (sp, nrows, nbits)
	register u_long *sp;
	register int nrows;
	int nbits;
{
	register int l = nbits % 32;
	register u_int j = ~mask[l];	/* # bits in last word */
	register int k;
	register int i = nbits / 32;	/* # full words to do */

	while (--nrows >= 0) {
		for (k = i; --k >= 0; )
			*sp++ = ~0;
		if (l)
			*sp++ = j;
	}
}

/*
 * Make "nc" copies of "nl" longs from "src" starting at "dst" 
 * (assumes that src < dst and src and dst don't overlap).
 */
static void
rowcopy(src, dst, nc, nl)
	register u_long *src;
	register u_long *dst;
	register int nc;
	register int nl;
{
	register u_long *sp;
	register int i;

	while (--nc >= 0) {
		sp = src;
		for (i = nl; --i >= 0; )
			*dst++ = *sp++;
	}
}

/*
 * Read bits from a bytestream.  Return the next 'nbits' bits
 * from pkin left adjusted in an 8 bit byte.  The 'buf' and 'bp'
 * temporaries are a buffer for the most recent byte read and
 * the number of unread bits in that byte respectively.
 */
#define READBITS(result, nbits, buf, bp) \
{ \
	if (bp) { \
		result = buf << (8 - bp); \
		if (nbits > bp) { \
			buf = _READ; \
			result |= (buf >> bp); \
			bp = 8 - (nbits - bp); \
		} else \
			bp = bp - nbits; \
	} else { \
		result = _READ; \
		bp = 8 - nbits; \
		if (bp) \
			buf = result; \
	} \
	if (8 - nbits) \
		result = (result >> (8 - nbits)) << (8 - nbits); \
}



/*
 * load a PK character into memory.
 */
void 
readpkbits (pkin, cp)
	register FILE *pkin;
	register struct char_entry *cp;
{
	int	cw;	/* character width */
	int	ch;	/* character height */
	u_int	dyn_f;	/* dynamic factor, part of a PK word. */
	u_long	*sp;
	int	rowsize;
	register int i, j;
	register int black, bits, bp, rc;
	int	rowp;

	if (cp->width == 0 || cp->height == 0) {
		cp->where.address.pixptr = (u_long *) 0;
		return;
	}
	cw = cp->width; ch = cp->height;

	/*
	 * what remains is the data for the image.  It can be packaged
	 * either as a run-encoding where successive values are the
	 * number of adjacent pixels to paint in the opposite color of
	 * the previous painting, or simply as a bitmap with no padding
	 * except (possibly) for the very last nybble to round up to a
	 * byte value.
	 *
	 * the data for the character is stored in successive bits with
	 * each new horizontal row at a long boundary.  See the PXL
	 * file format.
	 */
	rowsize = (cw + 31) / 32;
	sp = (u_long *)calloc (rowsize * cp->height, sizeof(u_long));
	if (! sp) {
		fprintf (stderr, "dvi2ps: readpkbits: calloc failed.\n");
		exit (1);
	}
	cp->where.address.pixptr = sp;

	/*
	 * grab the dyn_f out of the flag for this character.
	 */
	dyn_f = (cp->where.flags >> 4) & 0xf;
	/*
	 * the data returned by calloc is zeroed; we depend on that
	 * because we only turn on bits that are supposed to be
	 * black.
	 */
	if (dyn_f == 14) {
		/*
		 * we have a bitmap rather than a run-encoding.
		 */
		unsigned char *op = (unsigned char *)sp;
		int k;

		j = cw % 8;	/* extra bits in last byte */
		k = (32 - (cw % 32)) / 8;	/* padding bytes to add */
		bp = 0;
		while (--ch >= 0) {
			unsigned char tmp, buf;
			for (i = cw / 8; --i >= 0; ) {
				READBITS(tmp,8,buf,bp);
				*op++ = tmp;
			}
			if (j) {
				READBITS(tmp,j,buf,bp);
				*op++ = tmp;
			}
			for (i = k; --i >= 0; )
				*op++ = 0;
		}
		return;
	}

	bp = _pk_repeat = rc = rowp = 0;
	black = cp->where.flags & (1 << 3);
	bits = cw * ch;
	CLEAR_NYB;
	while (bp < bits) {
		GET_PACKED(j);

		if (_pk_repeat != 0) {
			rc = _pk_repeat;
			rowp = (bp / cw) * cw;
			_pk_repeat = 0;
			continue;
		}

		/*
		 * we have a run count in j.
		 */
		if (black) {
			register int k;

			k = bp % cw;		/* start bit offset */
			i = (bp + j) % cw;	/* ending bit offset */
			if (j <= cw && (k < i || i == 0)) {
				/* we're changing less than a row */
				FILL(bp, j);
			} else {
				/* fill any fragment in the current row */
				k = cw - k;
				FILL(bp, k);
				/* fill some number of full rows */
				FILLR(bp+k, (j - k) / cw);
				/* fill any fragment in the last row */
				FILL(bp+j-i, i);
			}
		}
		bp += j;

		/*
		 * if there's a repeat count and we hit the end of
		 * a row, do the copy.
		 */
		if (rc && (bp - rowp) >= cw) {
			i = rowp + cw;
			if (i != bp) {
				bcopy (ROWADDR(i), ROWADDR(i + rc*cw),
				       ((bp - i - 1) / cw + 1) * rowsize
				       * sizeof(*sp));
			}
			COPY(rowp, i, rc);
			bp += rc * cw;
			rc = 0;
		}
		black = !black;
	}
}
